package com.six_group.sse.seni.guicontrols.checkboxcombo;

import java.util.BitSet;
import java.util.Map;

import javafx.beans.InvalidationListener;
import javafx.beans.Observable;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.collections.ListChangeListener;
import javafx.collections.ObservableList;
import javafx.scene.control.MultipleSelectionModel;

import com.sun.javafx.collections.MappingChange;
import com.sun.javafx.collections.NonIterableChange;
import com.sun.javafx.scene.control.ReadOnlyUnbackedObservableList;

// not public API
// stolen as-is from ControlsFX -- tkcr1
abstract class CheckBitSetModelBase<T> extends MultipleSelectionModel<T> {

	/***********************************************************************
	 * * Internal properties * *
	 **********************************************************************/

	private final Map<T, BooleanProperty>					itemBooleanMap;

	private final BitSet									selectedIndices;
	private final ReadOnlyUnbackedObservableList<Integer>	selectedIndicesList;
	private final ReadOnlyUnbackedObservableList<T>			selectedItemsList;

	/***********************************************************************
	 * * Constructors * *
	 **********************************************************************/

	CheckBitSetModelBase(final Map<T, BooleanProperty> itemBooleanMap) {
		this.itemBooleanMap = itemBooleanMap;

		this.selectedIndices = new BitSet();

		this.selectedIndicesList = new ReadOnlyUnbackedObservableList<Integer>() {
			@Override
			public Integer get(int index) {
				if (index < 0 || index >= getItemCount())
					return -1;

				for (int pos = 0, val = selectedIndices.nextSetBit(0); val >= 0 || pos == index; pos++, val = selectedIndices.nextSetBit(val + 1)) {
					if (pos == index)
						return val;
				}

				return -1;
			}

			@Override
			public int size() {
				return selectedIndices.cardinality();
			}

			@Override
			public boolean contains(Object o) {
				if (o instanceof Number) {
					Number n = (Number) o;
					int index = n.intValue();

					return index >= 0 && index < selectedIndices.length() && selectedIndices.get(index);
				}

				return false;
			}
		};

		this.selectedItemsList = new ReadOnlyUnbackedObservableList<T>() {
			@Override
			public T get(int i) {
				int pos = selectedIndicesList.get(i);
				if (pos < 0 || pos >= getItemCount())
					return null;
				return getItem(pos);
			}

			@Override
			public int size() {
				return selectedIndices.cardinality();
			}
		};

		final MappingChange.Map<Integer, T> map = new MappingChange.Map<Integer, T>() {
			@Override
			public T map(Integer f) {
				return getItem(f);
			}
		};
		selectedIndicesList.addListener(new ListChangeListener<Integer>() {
			@Override
			public void onChanged(final Change<? extends Integer> c) {
				// when the selectedIndices ObservableList changes, we manually call
				// the observers of the selectedItems ObservableList.
				selectedItemsList.callObservers(new MappingChange<Integer, T>(c, map, selectedItemsList));
				c.reset();
			}
		});
	}

	/***********************************************************************
	 * * Abstract API * *
	 **********************************************************************/

	public abstract T getItem(int index);

	public abstract int getItemCount();

	public abstract int getItemIndex(T item);

	/***********************************************************************
	 * * Public selection API * *
	 **********************************************************************/

	/**
	 * Returns a read-only list of the currently selected indices in the
	 * CheckBox.
	 */
	@Override
	public ObservableList<Integer> getSelectedIndices() {
		return selectedIndicesList;
	}

	/**
	 * Returns a read-only list of the currently selected items in the CheckBox.
	 */
	@Override
	public ObservableList<T> getSelectedItems() {
		return selectedItemsList;
	}

	/** {@inheritDoc} */
	@Override
	public void selectAll() {
		for (int i = 0; i < getItemCount(); i++) {
			select(i);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void selectFirst() {
		select(0);
	}

	/** {@inheritDoc} */
	@Override
	public void selectIndices(int index, int... indices) {
		select(index);
		for (int i = 0; i < indices.length; i++) {
			select(indices[i]);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void selectLast() {
		select(getItemCount() - 1);
	}

	/** {@inheritDoc} */
	@Override
	public void clearAndSelect(int index) {
		clearSelection();
		select(index);
	}

	/** {@inheritDoc} */
	@Override
	public void clearSelection() {
		selectedIndices.clear();
	}

	/** {@inheritDoc} */
	@Override
	public void clearSelection(int index) {
		selectedIndices.clear(index);

		final int changeIndex = selectedIndicesList.indexOf(index);
		selectedIndicesList.callObservers(new NonIterableChange.SimpleRemovedChange<Integer>(changeIndex, changeIndex + 1, index, selectedIndicesList));
	}

	/** {@inheritDoc} */
	@Override
	public boolean isEmpty() {
		return selectedIndices.isEmpty();
	}

	/** {@inheritDoc} */
	@Override
	public boolean isSelected(int index) {
		return selectedIndices.get(index);
	}

	/** {@inheritDoc} */
	@Override
	public void select(int index) {
		if (index < 0 || index >= selectedIndices.size())
			;
		selectedIndices.set(index);

		final int changeIndex = selectedIndicesList.indexOf(index);
		selectedIndicesList.callObservers(new NonIterableChange.SimpleAddChange<Integer>(changeIndex, changeIndex + 1, selectedIndicesList));
	}

	/** {@inheritDoc} */
	@Override
	public void select(T item) {
		int index = getItemIndex(item);
		select(index);
	}

	/** {@inheritDoc} */
	@Override
	public void selectNext() {
		// no-op
	}

	/** {@inheritDoc} */
	@Override
	public void selectPrevious() {
		// no-op
	}

	/***********************************************************************
	 * * Private implementation * *
	 **********************************************************************/

	protected void updateMap() {
		// reset the map
		itemBooleanMap.clear();
		for (int i = 0; i < getItemCount(); i++) {
			final int index = i;
			final T item = getItem(index);

			final BooleanProperty booleanProperty = new SimpleBooleanProperty(item, "selected", false);
			itemBooleanMap.put(item, booleanProperty);

			// this is where we listen to changes to the boolean properties,
			// updating the selected indices list (and therefore indirectly
			// the selected items list) when the checkbox is toggled
			booleanProperty.addListener(new InvalidationListener() {
				@Override
				public void invalidated(Observable o) {
					final int changeIndex = selectedIndicesList.indexOf(index);

					if (booleanProperty.get()) {
						selectedIndices.set(index);
						selectedIndicesList.callObservers(new NonIterableChange.SimpleAddChange<Integer>(changeIndex, changeIndex + 1, selectedIndicesList));
					} else {
						selectedIndices.clear(index);
						selectedIndicesList.callObservers(new NonIterableChange.SimpleRemovedChange<Integer>(changeIndex, changeIndex + 1, index, selectedIndicesList));
					}
				}
			});
		}
	}
}