package com.six_group.sse.seni.guicontrols.checkboxcombo;

import java.util.HashMap;
import java.util.Map;

import javafx.beans.property.BooleanProperty;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.collections.FXCollections;
import javafx.collections.ListChangeListener;
import javafx.collections.ObservableList;
import javafx.scene.control.CheckBox;
import javafx.scene.control.Control;
import javafx.scene.control.MultipleSelectionModel;
import javafx.scene.control.Skin;

/**
 * A simple UI control that makes it possible to select zero or more items
 * within a ComboBox-like control. Each row item shows a {@link CheckBox}, and
 * the state of each row can be queried via the {@link #checkModelProperty()
 * check model}.
 * 
 * <h3>Screenshot</h3>
 * <p>
 * The following screenshot shows the CheckComboBox with some sample data:
 * 
 * <br/>
 * <img src="checkComboBox.png"/>
 * 
 * <h3>Code Example:</h3>
 * <p>
 * To create the CheckComboBox shown in the screenshot, simply do the following:
 * 
 * <pre>
 * {@code
 * // create the data to show in the CheckComboBox 
 * final ObservableList<String> strings = FXCollections.observableArrayList();
 * for (int i = 0; i <= 100; i++) {
 *     strings.add("Item " + i);
 * }
 * 
 * // Create the CheckComboBox with the data 
 * final CheckComboBox<String> checkComboBox = new CheckComboBox<String>(strings);
 * 
 * // and listen to the relevant events (e.g. when the selected indices or 
 * // selected items change).
 * checkComboBox.getCheckModel().getSelectedItems().addListener(new ListChangeListener<String>() {
 *     public void onChanged(ListChangeListener.Change<? extends String> c) {
 *         System.out.println(checkComboBox.getCheckModel().getSelectedItems());
 *     }
 * });}
 * }
 * </pre>
 *
 * @param <T>
 *            The type of the data in the ComboBox.
 */
// Override from ControlsFX, adds renderer implementations -- tkcr1
public class CheckBoxComboBox<T> extends Control {

	/**************************************************************************
	 * 
	 * Private fields
	 * 
	 **************************************************************************/

	private final ObservableList<T>			items;
	private final Map<T, BooleanProperty>	itemBooleanMap;

	/**************************************************************************
	 * 
	 * Constructors
	 * 
	 **************************************************************************/

	/**
	 * Creates a new CheckComboBox instance with an empty list of choices.
	 */
	public CheckBoxComboBox() {
		this(null);
	}

	/**
	 * Creates a new CheckComboBox instance with the given items available as
	 * choices.
	 * 
	 * @param items
	 *            The items to display within the CheckComboBox.
	 */
	public CheckBoxComboBox(final ObservableList<T> items) {
		final int initialSize = items == null ? 32 : items.size();

		this.itemBooleanMap = new HashMap<T, BooleanProperty>(initialSize);
		this.items = items == null ? FXCollections.<T> observableArrayList() : items;
		setCheckModel(new CheckComboBoxBitSetCheckModel<T>(this.items, itemBooleanMap));
	}

	/**************************************************************************
	 * 
	 * Public API
	 * 
	 **************************************************************************/

	/** {@inheritDoc} */
	@Override
	protected Skin<?> createDefaultSkin() {
		return new CheckBoxComboBoxSkin<>(this);
	}

	/**
	 * Represents the list of choices available to the user, from which they can
	 * select zero or more items.
	 */
	public ObservableList<T> getItems() {
		return items;
	}

	/**
	 * Returns the {@link BooleanProperty} for a given item index in the
	 * CheckComboBox. This is useful if you want to bind to the property.
	 */
	public BooleanProperty getItemBooleanProperty(int index) {
		if (index < 0 || index >= items.size())
			return null;
		return getItemBooleanProperty(getItems().get(index));
	}

	/**
	 * Returns the {@link BooleanProperty} for a given item in the
	 * CheckComboBox. This is useful if you want to bind to the property.
	 */
	public BooleanProperty getItemBooleanProperty(T item) {
		return itemBooleanMap.get(item);
	}

	/**************************************************************************
	 * 
	 * Properties
	 * 
	 **************************************************************************/

	// --- Check Model
	private ObjectProperty<MultipleSelectionModel<T>>	checkModel	= new SimpleObjectProperty<MultipleSelectionModel<T>>(this, "checkModel");

	/**
	 * Sets the 'check model' to be used in the CheckComboBox - this is the code
	 * that is responsible for representing the selected state of each
	 * {@link CheckBox} - that is, whether each {@link CheckBox} is checked or
	 * not (and not to be confused with the selection model concept, which is
	 * used in the ComboBox control to represent the selection state of each
	 * row)..
	 */
	public final void setCheckModel(MultipleSelectionModel<T> value) {
		checkModelProperty().set(value);
	}

	/**
	 * Returns the currently installed check model.
	 */
	public final MultipleSelectionModel<T> getCheckModel() {
		return checkModel == null ? null : checkModel.get();
	}

	/**
	 * The check model provides the API through which it is possible to check
	 * single or multiple items within a CheckComboBox, as well as inspect which
	 * items have been checked by the user. Note that it has a generic type that
	 * must match the type of the CheckComboBox itself.
	 */
	public final ObjectProperty<MultipleSelectionModel<T>> checkModelProperty() {
		return checkModel;
	}

	/**************************************************************************
	 * 
	 * Implementation
	 * 
	 **************************************************************************/

	// tkcr1 - added all below renderer	
	private ICheckBoxComboBoxCellRenderer<T>	_cellRenderer;

	public ICheckBoxComboBoxCellRenderer<T> getCellRenderer() {
		return _cellRenderer;
	}

	public void setCellRenderer(ICheckBoxComboBoxCellRenderer<T> cellRenderer) {
		_cellRenderer = cellRenderer;
	}

	private ICheckBoxComboBoxTitleRenderer<T>	_titleRenderer;

	public ICheckBoxComboBoxTitleRenderer<T> getTitleRenderer() {
		return _titleRenderer;
	}

	public void setTitleRenderer(ICheckBoxComboBoxTitleRenderer<T> titleRenderer) {
		_titleRenderer = titleRenderer;
	}

	/**************************************************************************
	 * 
	 * Support classes
	 * 
	 **************************************************************************/

	private static class CheckComboBoxBitSetCheckModel<T> extends CheckBitSetModelBase<T> {

		/***********************************************************************
		 * * Internal properties * *
		 **********************************************************************/

		private final ObservableList<T>	items;

		/***********************************************************************
		 * * Constructors * *
		 **********************************************************************/

		CheckComboBoxBitSetCheckModel(final ObservableList<T> items, final Map<T, BooleanProperty> itemBooleanMap) {
			super(itemBooleanMap);

			this.items = items;
			this.items.addListener(new ListChangeListener<T>() {
				@Override
				public void onChanged(Change<? extends T> c) {
					updateMap();
				}
			});

			updateMap();
		}

		/***********************************************************************
		 * * Implementing abstract API * *
		 **********************************************************************/

		@Override
		public T getItem(int index) {
			return items.get(index);
		}

		@Override
		public int getItemCount() {
			return items.size();
		}

		@Override
		public int getItemIndex(T item) {
			return items.indexOf(item);
		}
	}

}
