package dsfxcontrols;

import java.awt.Color;
import java.awt.Rectangle;
import java.util.logging.Logger;
import javafx.application.Application;
import javafx.beans.property.DoubleProperty;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleDoubleProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.collections.ObservableList;
import javafx.scene.Node;
import javafx.scene.Scene;
import javafx.scene.control.Button;
import javafx.scene.control.ColorPicker;
import javafx.scene.layout.VBox;
import javafx.stage.Stage;
import javafx.util.Callback;
import org.controlsfx.control.PropertySheet;
import org.controlsfx.property.BeanPropertyUtils;
import org.controlsfx.property.editor.PropertyEditor;

/**
 *
 * SSCE to demonstrate:
 * <ul>
 * <li>Property editors added with setPropertyEditorFactory are not invoked when
 * a property values is altered. Demonstrate by pressing the "New Editor" button
 * twice to create two PropertySheet editors, then try changing the AWT color in
 * one editor and observe that it fails to update in the other editor.
 * <li>PropertySheets fail stop listening to the properties (and thus keep them
 * in memory) when the PropertySheet is closed. Demonstrate by creating multiple
 * editors, edit the value of property "ukp" and observe how many debug messages
 * are displayed for "getEuro".. one for each open PropertyEditorSheet. Delete
 * all but one of the PropertySheetEditors and click the "Garbage Collect"
 * button. edit the value of property "ukp" and observe how many debug messages
 * are displayed for "getEuro" and see that a call is till made for all the
 * deleted PropertySheetEditors.
 * <li>The formatting used for the default property editor for values of type
 * Double makes it difficult to edit try selecting the whole text and then
 * entering the value "40".
 * </ul>
 *
 * @author Michel Ellis
 */
public class TestExtendedPropertySheet2 extends Application {

    private final static String CLASS_NAME = TestExtendedPropertySheet2.class.getName();
    final static Logger LOGGER = Logger.getLogger(CLASS_NAME);

    MyBean myBean = new MyBean();

    @Override
    public void start(Stage propertySheetStage) {
        final VBox settingsEditor = new VBox();
        final Button newEditorButton = new Button("New Editor");
        newEditorButton.setOnAction(e -> handleCreateEditor(settingsEditor));
        final Button gcButton = new Button("Garbage Collect");
        gcButton.setOnAction(e -> System.gc());

        settingsEditor.getChildren().addAll(newEditorButton, gcButton);

        Scene scene = new Scene(settingsEditor);
        propertySheetStage.setScene(scene);
        propertySheetStage.show();
    }

    int initialPos = 100;

    /**
     * Create a PropertySheetEditor that has additional support for
     * java.awt.Color
     *
     * @param root Parent window
     */
    void handleCreateEditor(Node root) {
        ObservableList<PropertySheet.Item> props = BeanPropertyUtils.getProperties(myBean);
        PropertySheet propertySheet = new PropertySheet(props);

        // Get the default property Sheet editor
        final Callback<PropertySheet.Item, PropertyEditor<?>> defaultPropertyEditorFactory = propertySheet.getPropertyEditorFactory();

        // Insert our AwtColorEditor for editing java.awt.Color porperties
        propertySheet.setPropertyEditorFactory(
                item -> (item.getType() == java.awt.Color.class)
                        ? new AwtColorEditor(item)
                        : defaultPropertyEditorFactory.call(item)
        );

        Scene scene = new Scene(propertySheet);

        final Stage propertySheetStage = new Stage();
        propertySheetStage.setTitle("Property Editor");
        propertySheetStage.setScene(scene);
        propertySheetStage.initOwner(root.getScene().getWindow());
        propertySheetStage.setX(initialPos);
        propertySheetStage.setY(initialPos);
        initialPos += 30;
        propertySheetStage.show();
    }

    public static void main(String[] args) {
        launch(args);
    }

    /**
     * A Bean with some FX properties of various types. Note the UKP and Euro
     * property are linked. Setting the value of UKP also sets the Euro property
     * and visa-versa.
     *
     * Beans Means Heinz https://www.youtube.com/watch?v=RGhWcPRHMAI
     */
    public class MyBean {

        public MyBean() {
        }

        final private SimpleDoubleProperty ukp = new SimpleDoubleProperty(this, "ukp", 0.0);

        public double getUkp() {
            LOGGER.severe("getUkp()");
            return ukp.get();
        }

        public void setUkp(double ukp) {
            this.ukp.set(ukp);
            this.euro.set(ukp * 1.6);
        }

        public DoubleProperty ukpProperty() {
            LOGGER.severe("ukpProperty()");
            return ukp;
        }

        final private SimpleDoubleProperty euro = new SimpleDoubleProperty(this, "euro", 0.0);

        public double getEuro() {
            LOGGER.severe("getEuro()");
            return euro.get();
        }

        public void setEuro(double euro) {
            this.euro.set(euro);
            this.ukp.set(euro / 1.6);
        }

        public DoubleProperty euroProperty() {
            LOGGER.severe("euroProperty()");
            return euro;
        }

        final private ObjectProperty<Rectangle> ROIProperty = new SimpleObjectProperty<>(this, "ROI", new Rectangle(1, 2, 3, 4));

        public Rectangle getROI() {
            LOGGER.severe("getROI()");
            return ROIProperty.get();
        }

        public void setROI(Rectangle ROI) {
            this.ROIProperty.set(ROI);
        }

        public ObjectProperty<Rectangle> ROIProperty() {
            return ROIProperty;
        }

        final private ObjectProperty<Color> AwtColorProperty = new SimpleObjectProperty<>(this, "AwtColor", Color.RED);

        public Color getAwtColor() {
            LOGGER.severe("getAwtColor()");
            return AwtColorProperty.get();
        }

        public void setAwtColor(Color color) {
            this.AwtColorProperty.set(color);
        }

        public ObjectProperty<Color> AwtColorProperty() {
            return AwtColorProperty;
        }

    }
}

/**
 * Property Editor for java.awt.Color properties.
 */
final class AwtColorEditor implements PropertyEditor<java.awt.Color> {

    PropertySheet.Item item;
    ColorPicker colorPicker;

    AwtColorEditor(PropertySheet.Item item) {
        this.item = item;
        colorPicker = new ColorPicker();
        colorPicker.setOnAction(ae -> item.setValue(getValue()));
    }

    @Override
    public Node getEditor() {
        return colorPicker;
    }

    @Override
    public java.awt.Color getValue() {
        javafx.scene.paint.Color c = colorPicker.getValue();
        return new java.awt.Color((float) c.getRed(), (float) c.getGreen(), (float) c.getBlue());
    }

    @Override
    public void setValue(java.awt.Color d) {
        if (d != null) {
            colorPicker.setValue(javafx.scene.paint.Color.rgb(d.getRed(), d.getGreen(), d.getBlue()));
        }
    }
}
