package bugs.controlsfx.propertysheet;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.text.ParseException;
import java.util.Optional;
import java.util.function.UnaryOperator;

import org.controlsfx.control.PropertySheet;
import org.controlsfx.control.PropertySheet.Item;

import javafx.application.Application;
import javafx.beans.value.ObservableValue;
import javafx.geometry.Pos;
import javafx.scene.Scene;
import javafx.scene.control.Label;
import javafx.scene.control.TextField;
import javafx.scene.control.TextFormatter;
import javafx.scene.control.TextFormatter.Change;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.HBox;
import javafx.scene.layout.VBox;
import javafx.stage.Stage;
import javafx.util.StringConverter;

/**
 * Problems with numerical fields in PropertySheet.
 * 
 * This application shows various problems when editing numerical fields in a property sheet.
 * In order to show all these problems you have to start the application with the german
 * locale via "-Duser.language=de". The application shows a properly configured JavaFX TextField at the
 * top and a numerical field of a property sheet at the bottom.
 * 1. The correct display of the example number should be "3,14" as shown in the TextField and not "3.14".
 * The property sheet completely ignores the default locale of the application.
 * 2. Now try to change this number to "123,45" in the property sheet. When the field gets the focus
 * everything is selected and you start punching in the digits 1 2 3 , 4 5 . The number you end up
 * with is "1.02345" which is of course wrong. This is because the formatter always appends the decimal
 * separator even if it is not needed but is then not able to cope with it correctly. Try the same in the
 * TextField and it works.
 * 3. To my opinion numerical fields should be right-aligned by default and not left-aligned.
 * The problems described here could be avoided by properly configuring the numerical text field
 * used inside the property sheet by making use of the DecimalFormat which gets it all right by default.
 * You only have to take care of intermediate states inside the filter which may not yet be
 * parsable numbers.
 */
public class PropertySheetLocalizationBug extends Application {
	
	private final static DecimalFormat df = (DecimalFormat)NumberFormat.getNumberInstance();
	
	public final static StringConverter<Double> doubleStringConverter = new StringConverter<Double>() {
		@Override
		public String toString(Double object) {
			if (object != null) {
				return df.format(object);
			} else {
				return "";
			}
		}

		@Override
		public Double fromString(String string) {
			try {
				return df.parse(string).doubleValue();
			} catch (ParseException e) {
				throw new IllegalArgumentException("Cannot parse '" + string + "'");
			}
		}
	};
	
	public final static UnaryOperator<TextFormatter.Change> doubleFilter = new UnaryOperator<TextFormatter.Change>() {

		@Override
		public Change apply(Change t) {
			String cnt = t.getControlNewText();
			try {				
				df.parse(cnt);
			} catch (ParseException e) {
				if (!cnt.trim().equals(cnt)) return null;
				if (cnt.isEmpty()) return t;
				if (cnt.equals(String.valueOf(df.getDecimalFormatSymbols().getMinusSign()))) return t;
				// This is not yet complete!
								
				return null;
			}
			return t;
		}
		
	};
	
    class DoublePropertyItem implements Item {

    	private String name;
        private Double value;

        public DoublePropertyItem(String name, double value) {
        	this.name = name;
        	this.value = value;
        }

        @Override
        public Class<?> getType() {
            return Double.class;
        }

        @Override
        public String getCategory() {
            return "doubles";
        }

        @Override
        public String getName() {
            return name;
        }

        @Override
        public String getDescription() {
            return null;
        }

        @Override
        public Object getValue() {
            return value;
        }

        @Override
        public void setValue(Object value) {
            this.value = (Double)value;
        }

        @Override
        public Optional<ObservableValue<? extends Object>> getObservableValue() {
            return Optional.empty();
        }

    }
	
	private BorderPane root;
	
	@Override
	public void start(final Stage stage) throws Exception {
        stage.setTitle(PropertySheetLocalizationBug.class.getSimpleName());
        
        root = new BorderPane();
        Scene scene = new Scene(root, 400, 200);
        
        HBox hbox = new HBox();
        hbox.setAlignment(Pos.CENTER);
        
        VBox vbox = new VBox();
        vbox.setAlignment(Pos.CENTER);
        hbox.getChildren().add(vbox);
        
        vbox.getChildren().add(new Label("JavaFX TextField with TextFormatter."));
                
        TextFormatter<Double> numericTextFormatter = new TextFormatter<>(doubleStringConverter, null, doubleFilter);
        TextField numericTextField = new TextField();
        numericTextField.setTextFormatter(numericTextFormatter);
        numericTextField.setAlignment(Pos.BASELINE_RIGHT);
        vbox.getChildren().add(numericTextField);
        numericTextFormatter.setValue(3.14);
        
        vbox.getChildren().add(new Label(""));
        
        vbox.getChildren().add(new Label("ControlsFX PropertySheet."));
        
        PropertySheet propertySheet = new PropertySheet();
        propertySheet.getItems().add(new DoublePropertyItem("Double item", 3.14));
        vbox.getChildren().add(propertySheet);
        
        root.setCenter(hbox);
        
        stage.setScene(scene);
        stage.show();
	}

	public static void main(String[] args) {
		launch(args);
	}
	
}
