import javafx.application.Application;
import javafx.geometry.Insets;
import javafx.geometry.Pos;
import javafx.scene.Node;
import javafx.scene.Scene;
import javafx.scene.control.Dialog;
import javafx.scene.control.TextField;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.VBox;
import javafx.stage.Stage;
import org.controlsfx.control.decoration.Decorator;
import org.controlsfx.control.decoration.GraphicDecoration;

public class TestGraphicDecorationListener extends Application {

   /*
   This app demonstrates that this fix ...

   https://bitbucket.org/controlsfx/controlsfx/issues/495/decorator-always-in-top-left-corner

   ... does not seem to work, that it never gets called because it fails its zero values
   check for targetWidth and targetHeight.  Line 150+ in GraphicDecoration.

   This app will create a simle FX app, displaying two TextFields, both of them having
   a decoration added to them, but one of the will rely on the listener inside of
   GraphicDecoration to detect/handle target Node's bounds changes, while the other will
   have our own manually added listener for target Node's bound changes added to it.

   Note that for the first TextField, the decoration/image will NOT move, as you resize the
   Stage horizontally, while the second TextField, will.  The first relys on the listener being
   added by GraphicDecoration, while the second one uses our own manually added listener.

   This coded/test with version 1.8.0_152 of Oracle JDK, and version 8.40.14 of ControlsFX.
    */

   @Override
   public void start( final Stage primaryStage ) throws Exception {

      final BorderPane pane = new BorderPane();
      final Scene scene = new Scene( pane, 500, 500 );
      primaryStage.setScene( scene );

      final TextField textField1 = new TextField( "Using GraphicDecoration's listener." );
      ////////////////////////////////////////////
      setDecoration( textField1, false ); // <-- Look at this method!
      ////////////////////////////////////////////

      final TextField textField2 = new TextField( "Adding our own listener manually!" );
      ////////////////////////////////////////////
      setDecoration( textField2, true ); // <-- Look at this method!
      ////////////////////////////////////////////

      final VBox vBox = new VBox( textField1, textField2 );
      vBox.setSpacing( 50 );
      vBox.setPadding( new Insets( 30, 30, 30, 30 ) );

      pane.setCenter( vBox );

      primaryStage.show();
   }

   /*
    * All of the 'wiring' is done in this method.
    */
   private void setDecoration( final Node node, final boolean setListener ) {

      System.out.println(System.getProperty("java.version"));

      // Create the image that will be viewed from the decorator ...
      final ImageView imageView = new ImageView( new Image( Dialog.class.getResource(
            // "dialog-information.png" "dialog-confirm.png" "dialog-warning.png" "dialog-error.png"
            // "dialog-more-details.png" "dialog-fewer-details.png"
            "/com/sun/javafx/scene/control/skin/modena/dialog-information.png"
      ).toExternalForm() ) );

      // Create the decorator, using the image ...
      final GraphicDecoration decoration = new GraphicDecoration( imageView, Pos.CENTER_RIGHT );

      // Add the decoration to its Node ...
      // Platform.runLater( () -> Decorator.addDecoration( node, decoration ) );
      Decorator.addDecoration( node, decoration );


      /* //////////////////////////////////////////////////////////////////////////////////////////////
      I would not expect to have to do the below code, as GraphicDecoration adds its own listener to
      detect when the Node that it is displaying an image for changes its bounds (line 150ish). However
      it does not seem to get called, because the checks for (targetWidth <= 0 && targetHeight <= 0)
      fails with non-zero values.  The below if statement allows us to either add or not add our own
      bounds listener, to demonstrate that the decoration will move in relation to its owner Node when
      the owner Node changes its bounds, but only if we add our own, proving that GraphicDecoration's
      listener is not being fired off.
       */ ////////////////////////////////////////////////////////////////////////////////////////////

      // Add (or not) our own listener to the Node for when it changes it bounds, to update decoration ...
      if (setListener) {
         // GraphicDecoration is not adding a listener to the owner
         // for owner bounds changes!  Line 150 class GraphicDecoration.
         // https://bitbucket.org/controlsfx/controlsfx/issues/495/decorator-always-in-top-left-corner
         node.layoutBoundsProperty().addListener( ( observable, oldValue, newValue ) -> {
            // updateGraphicPosition( node ); // Wish this wasn't private! Then again, shouldn't need to call it.
            // Calling "GraphicDecoration.updateGraphicPosition(node)" indirectly ...
            // decoration.removeDecoration( node ); // This doesn't seem necessary, as 'apply' does dup check.
            decoration.applyDecoration( node );
         } );
      }

   }

}
