package eu.xxxxx.ting.gui.components.spectable.cells;

import org.controlsfx.control.spreadsheet.SpreadsheetCell;
import org.controlsfx.control.spreadsheet.SpreadsheetCellBase;
import org.controlsfx.control.spreadsheet.SpreadsheetCellType;

import de.di.app.errors.Arg;

/**
 * Describes the general logic that CellHandlers should implement.<br>
 * <br>
 * {@code CellHandlers} are extensions to the standard {@link SpreadsheetCellType}s with additional logic and behaviour.
 */
public abstract class CellHandler
{
  /************************************* CONSTANTS *******************************************/
  
  public final static SpreadsheetCellFactory DEFAULT_FACTORY = SpreadsheetCellBase::new;
  
  /********************************** INTERNAL TYPES *****************************************/
  
  /**
   * Builds a {@link SpreadsheetCell}. Necessary for applications that need special implementations of
   * {@code SpreadsheetCellBase}. <br>
   * <br>
   * The default implementation is {@link CellHandler#DEFAULT_FACTORY}.
   */
  public interface SpreadsheetCellFactory
  {
    /** Creates a new instance of {@link SpreadsheetCell}. */
    public SpreadsheetCell createCell(final int row, final int column, final int rowSpan, final int columnSpan, final SpreadsheetCellType<?> type);
    
    /** Creates a new instance of {@link SpreadsheetCell}. Cell type is {@link SpreadsheetCellType.OBJECT}. */
    default SpreadsheetCell createCell(final int row, final int column, final int rowSpan, final int columnSpan)
    {
      return createCell(row, column, rowSpan, columnSpan, SpreadsheetCellType.OBJECT);
    }
  }
  
  /*************************************** FIELDS ********************************************/
  
  private int                    _row, _column;
  private int                    _rowSpan, _columnSpan;
  private SpreadsheetCellFactory _cellFactory;
  private boolean                _isInitiallyEditable = true;
  private String                 _initialCellStyle    = "";
  
  
  private SpreadsheetCell _associatedCell = null;
  
  /************************************ CONSTRUCTORS *****************************************/
  
  /**
   * Constructs a {@link DoubleCellHandler}. The underlying cell will be a {@link SpreadsheetCellBase}.
   *
   * @param row Row number
   * @param column Column number
   * @param rowSpan Row span (1 is normal)
   * @param columnSpan Column span (1 is normal)
   */
  public CellHandler(final int row, final int column, final int rowSpan, final int columnSpan)
  {
    this(row, column, rowSpan, columnSpan, DEFAULT_FACTORY);
  }
  
  /**
   * Constructs a {@link DoubleCellHandler}.
   *
   * @param row Row number
   * @param column Column number
   * @param rowSpan Row span (1 is normal)
   * @param columnSpan Column span (1 is normal)
   * @param cellFactory Provides cells of the desired type. Default implementation {@link CellHandler#DEFAULT_FACTORY}
   *          returns {@link SpreadsheetCellBase}. <b>Must not be <code>null</code>!</b>
   */
  public CellHandler(final int row, final int column, final int rowSpan, final int columnSpan, SpreadsheetCellFactory cellFactory)
  {
    Arg.notNull(this, cellFactory, "cellFactory");
    _row = row;
    _column = column;
    _rowSpan = rowSpan;
    _columnSpan = columnSpan;
    _cellFactory = cellFactory;
  }
  
  
  /************************************ GETTERS/SETTERS **************************************/
  
  public int getRow()
  {
    return _row;
  }
  
  public CellHandler setRow(int row)
  {
    _row = row;
    return this;
  }
  
  public int getColumn()
  {
    return _column;
  }
  
  public CellHandler setColumn(int column)
  {
    _column = column;
    return this;
  }
  
  public int getRowSpan()
  {
    return _rowSpan;
  }
  
  /**
   * Sets the row span.
   *
   * @param rowSpan The row span. The typical value is 1.
   */
  public CellHandler setRowSpan(int rowSpan)
  {
    _rowSpan = rowSpan;
    return this;
  }
  
  public int getColumnSpan()
  {
    return _columnSpan;
  }
  
  /**
   * Sets the column span.
   *
   * @param columnSpan The column span. The typical value is 1.
   */
  public CellHandler setColumnSpan(int columnSpan)
  {
    _columnSpan = columnSpan;
    return this;
  }
  
  public boolean isInitiallyEditable()
  {
    return _isInitiallyEditable;
  }
  
  /**
   * Sets if the cell will initially be editable.
   */
  public CellHandler setInitiallyEditable(boolean isInitiallyEditable)
  {
    _isInitiallyEditable = isInitiallyEditable;
    return this;
  }
  
  public String getInitialCellStyle()
  {
    return _initialCellStyle;
  }
  
  /**
   * Sets the initial cell style.
   */
  public CellHandler setInitialCellStyle(String initialCellStyle)
  {
    _initialCellStyle = initialCellStyle;
    return this;
  }
  
  /** Returns the cell factory which will be used to create the {@link SpreadsheetCell} instances. */
  protected SpreadsheetCellFactory getCellFactory()
  {
    return _cellFactory;
  }
  
  
  
  /************************************* GENERAL LOGIC ***************************************/
  
  /**
   * Builds a {@link SpreadsheetCell} with the given settings.<br>
   * <br>
   * If called multiple times, <u>always the same row instance</u> will be returned.
   *
   * @return Returns the {@code SpreadsheetCell} with the desired behaviour.
   */
  public final SpreadsheetCell getCell()
  {
    if ( _associatedCell == null )
    {
      _associatedCell = buildCell();
    }
    return _associatedCell;
  }
  
  /**
   * Builds a {@link SpreadsheetCell} with the given settings.
   *
   * @return Returns the {@code SpreadsheetCell} with the desired behaviour.
   */
  protected abstract SpreadsheetCell buildCell();
  
  
  /*******************************************************************************************/
  
  
}
