/*
 * Erstellt: 25.06.2018
 * Durch   : rvo
 * Geändert: $Date: 2018-11-01 09:45:56 +0100 (Thu, 01 Nov 2018) $
 * Durch   : $Author: rvo $
 * Revision: $Rev: 1370 $
 */
package eu.xxxxx.ting.gui.components.spectable.cells;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.HashSet;
import java.util.Locale;
import java.util.Set;
import java.util.function.BiFunction;

import org.controlsfx.control.spreadsheet.SpreadsheetCell;
import org.controlsfx.control.spreadsheet.SpreadsheetCellBase;
import org.controlsfx.control.spreadsheet.SpreadsheetCellEditor;
import org.controlsfx.control.spreadsheet.SpreadsheetCellType;
import org.controlsfx.control.spreadsheet.SpreadsheetView;
import org.controlsfx.control.spreadsheet.StringConverterWithFormat;
import org.controlsfx.validation.Severity;
import org.controlsfx.validation.ValidationSupport;
import org.controlsfx.validation.Validator;
import org.controlsfx.validation.decoration.StyleClassValidationDecoration;

import de.di.app.util.StringUtil;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.event.EventHandler;
import javafx.geometry.Insets;
import javafx.geometry.Pos;
import javafx.scene.Node;
import javafx.scene.control.IndexRange;
import javafx.scene.control.Label;
import javafx.scene.control.TextField;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;
import javafx.scene.input.MouseButton;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.ColumnConstraints;
import javafx.scene.layout.GridPane;
import javafx.scene.layout.StackPane;

/**
 * Extends the standard {@link SpreadsheetCellType}s by additional logic and behaviour.<br>
 * <br>
 * Each instance should be used to create <u>exactly one</u> instance of {@link SpreadsheetCell}, using
 * {@link #getCell()}.
 *
 * @author <a href="mailto:robert.voelckner@dmos2002.de">Robert Voelckner</a>
 * @version $Rev: 1370 $
 *
 */
public class MinMaxDoubleCellHandler extends CellHandler
{
  /************************************* CONSTANTS *******************************************/
  
  public static final String DEFAULT_STRING_FORMAT = "#.###############";
  
  
  /********************************** SUPPORTING TYPES ***************************************/
  
  /** Identifies the regarding field within callbacks */
  public static enum WhichField
  {
    MIN, MAX
  }
  
  /** Defines the callback interface for committing entered values */
  public static interface CommitValueCallback
  {
    void commitValue(SpreadsheetCell cell, WhichField whichField, double convertedValue, String originallyEnteredString);
  }
  
  /** Converter interface for user interputs */
  public interface ValueConverter
  {
    /** Will be called when a value shall be output. */
    String toString(WhichField whichField, double value, String format);
    
    /**
     * Will be called as soon as the user entered a valid value. Performs the translation to the internal data
     * representation.
     *
     * @param valueString Typically corresponds to the {@link String} value entered by the user.
     * @return Returns the value's internal representation. Will, internally, be assigned to
     *         {@link SpreadsheetCell#setItem(Object)}. Should be {@code null} if value is not parsable.
     */
    Double convertValue(WhichField whichField, String valueString);
    
  }
  
  
  /*************************************** FIELDS ********************************************/
  
  private double         _value1, _value2;
  private Set<Character> _allowedExtraSymbols    = null;
  private boolean        _isEditable             = true;
  private boolean        _isHighlightedIfInvalid = false;
  
  private CommitValueCallback                     _commitValueCallback = null;
  private BiFunction<WhichField, String, Boolean> _checkValueCallback;
  
  private ValueConverter _valueConverter = DEFAULT_CONVERTER; // Must not be null!
  
  // The following fields refer to components of the associated cell
  private Label     _valueLabel1  = null, _valueLabel2 = null;
  private StackPane _minValuePane = null, _maxValuePane = null;
  
  private EditorHandler _editorHandler = null;
  
  /************************************ CONSTRUCTORS *****************************************/
  
  /**
   * Constructs a {@link DoubleCellHandler}. The underlying cell will be a {@link SpreadsheetCellBase}.
   *
   * @param row Row number
   * @param column Column number
   * @param rowSpan Row span (1 is normal)
   * @param columnSpan Column span (1 is normal)
   * @param value The value that shall be displayed in the 'min' cell.
   * @param value The value that shall be displayed in the 'max' cell.
   */
  public MinMaxDoubleCellHandler(final int row, final int column, final int rowSpan, final int columnSpan, double initialMinValue, double initialMaxValue)
  {
    this(row, column, rowSpan, columnSpan, initialMinValue, initialMaxValue, null);
  }
  
  /**
   * Constructs a {@link DoubleCellHandler}. The underlying cell will be a {@link SpreadsheetCellBase}.
   *
   * @param row Row number
   * @param column Column number
   * @param rowSpan Row span (1 is normal)
   * @param columnSpan Column span (1 is normal)
   * @param value The value that shall be displayed in the 'min' cell.
   * @param value The value that shall be displayed in the 'max' cell.
   * @param commitValueCallback The callback function that is called when an entered (and successfully verified) value
   *          should be committed be the application. May be {@code null}.
   */
  public MinMaxDoubleCellHandler(final int row, final int column, final int rowSpan, final int columnSpan, double initialMinValue, double initialMaxValue, CommitValueCallback commitValueCallback)
  {
    this(row, column, rowSpan, columnSpan, initialMinValue, initialMaxValue, commitValueCallback, CellHandler.DEFAULT_FACTORY);
  }
  
  /**
   * Constructs a {@link DoubleCellHandler}.
   *
   * @param row Row number
   * @param column Column number
   * @param rowSpan Row span (1 is normal)
   * @param columnSpan Column span (1 is normal)
   * @param value The value that shall be displayed in the 'min' cell.
   * @param value The value that shall be displayed in the 'max' cell.
   * @param commitValueCallback The callback function that is called when an entered (and successfully verified) value
   *          should be committed be the application. May be {@code null}.
   * @param cellFactory Provides cells of the desired type. Default implementation {@link CellHandler#DEFAULT_FACTORY}
   *          returns {@link SpreadsheetCellBase}. <b>Must not be <code>null</code>!</b>
   */
  public MinMaxDoubleCellHandler(
    final int row,
    final int column,
    final int rowSpan,
    final int columnSpan,
    double initialMinValue,
    double initialMaxValue,
    CommitValueCallback commitValueCallback,
    SpreadsheetCellFactory cellFactory)
  {
    super(row, column, rowSpan, columnSpan, cellFactory);
    _value1 = initialMinValue;
    _value2 = initialMaxValue;
    setCommitValueCallback(commitValueCallback);
  }
  
  /************************************ GETTERS/SETTERS **************************************/
  
  /**
   * Sets the (optional) {@code CheckValueCallback} which will be called frequently while typing and once when the user
   * hits Enter, but before the text input field closes. May be used to check value ranges, string formatting, etc.<br>
   * <br>
   * This function will only be called if the value is parsable, using the given {@code ValueConverter}.
   *
   * @param checkValueCallback The {@code CommitValueCallback}. May be {@code null}.
   */
  public MinMaxDoubleCellHandler setCheckValueCallback(BiFunction<WhichField, String, Boolean> checkValueCallback)
  {
    _checkValueCallback = checkValueCallback;
    return this;
  }
  
  /**
   * Sets the (optional) {@code CommitValueCallback} which will be called after the text input field closed. The value
   * was already successfully checked and shall now be committed to the application's data basis, if necessary.
   *
   * @param commitValueCallback The {@code CommitValueCallback}. May be {@code null}.
   */
  public MinMaxDoubleCellHandler setCommitValueCallback(CommitValueCallback commitValueCallback)
  {
    _commitValueCallback = commitValueCallback;
    return this;
  }
  
  /**
   * Overrides the default {@link ValueConverter<Double>}. May be useful in connection with
   * {@link #setAllowedExtraSybols(Set)}.
   *
   * @param valueConverter The alternative {@code ValueConverter}. If {@code null}, the default {@code ValueConverter}
   *          will be applied.
   */
  public MinMaxDoubleCellHandler setValueConverter(ValueConverter valueConverter)
  {
    if ( valueConverter != null )
      _valueConverter = valueConverter;
    else
      _valueConverter = DEFAULT_CONVERTER;
    return this;
  }
  
  /**
   * Sets the additionally allowed symbols the user may enter - besides {@code [0-9], '.', ','}.<br>
   * <br>
   * When adding extra symbols, inserting a custom {@code ValueConverter<Double>} using
   * {@link #setValueConverter(ValueConverter)} is recommended.
   *
   * @param allowedExtraSymbols These symbols will additionally be allowed to enter. May be {@code null}.
   */
  public MinMaxDoubleCellHandler setAllowedExtraSybols(Set<Character> allowedExtraSymbols)
  {
    _allowedExtraSymbols = allowedExtraSymbols;
    return this;
  }
  
  /**
   * Sets if the cell contents are editable.<br>
   * <br>
   * For this type of {@code CellHandler}, {@link super#setInitiallyEditable(boolean)} is useless.
   */
  public MinMaxDoubleCellHandler setEditable(boolean isEditable)
  {
    _isEditable = isEditable;
    return this;
  }
  
  /**
   * Returns if the cell contents are editable.<br>
   * <br>
   * For this type of {@code CellHandler}, {@link super#isInitiallyEditable(boolean)} is useless.
   */
  public boolean isEditable()
  {
    return _isEditable;
  }
  
  /**
   * Denotes if invalid user inputs shall be highlighted using {@link ValidationSupport}. Invalid user inputs will be
   * determined using {@code ValueConverter} and {@code CheckValueCallback}.<br>
   * <br>
   * <br>
   * The highlighting scheme may be included into the style sheet as follows:
   * 
   * <pre>
   * .error {
   *    -fx-effect: innershadow(three-pass-box, darkred, 7, 0, 0, 0);
   * }
   * </pre>
   *
   * @param isHightlightedIfInvalid
   */
  public void setHightlightedIfInvalid(boolean isHightlightedIfInvalid)
  {
    _isHighlightedIfInvalid = isHightlightedIfInvalid;
  }
  
  
  /** {@inheritDoc} */
  @Override
  protected SpreadsheetCell buildCell()
  {
    _isEditable = isInitiallyEditable();
    
    //    SpreadsheetCell cell = getCellFactory().createCell(getRow(), getColumn(), getRowSpan(), getColumnSpan(), new MinMaxDoubleCellType());
    //    cell.setEditable(true);
    SpreadsheetCell cell = getCellFactory().createCell(getRow(), getColumn(), getRowSpan(), getColumnSpan());
    cell.setEditable(false);
    cell.setItem(null);
    cell.setGraphic(createCellComponents());
    cell.setStyle(getInitialCellStyle());
    cell.setFormat(DEFAULT_STRING_FORMAT);
    return cell;
  }
  
  
  /*********************************** PUBLIC LOGIC ******************************************/
  
  /**
   * Updates the cell's labels.
   */
  public void updateLabels()
  {
    SpreadsheetCell associatedCell = getCell();
    
    _valueLabel1.setText(_valueConverter.toString(WhichField.MIN, _value1, associatedCell.getFormat()));
    _valueLabel2.setText(_valueConverter.toString(WhichField.MAX, _value2, associatedCell.getFormat()));
  }
  
  /**
   * Opens the editor for one of both fields. If the cell is not editable or the editor is already open, nothing
   * happens.
   *
   * @param whichField Denotes the field (min/max) the editor shall be opened for.
   */
  public void requestOpenEditor(WhichField whichField)
  {
    if ( !_isEditable )
      return;
    if ( _editorHandler != null && _editorHandler.isCurrentlyOpen() )
      requestCloseEditor(false);
    
    if ( whichField == WhichField.MIN )
      _editorHandler = new EditorHandler(_minValuePane, _value1, whichField);
    else
      _editorHandler = new EditorHandler(_maxValuePane, _value2, whichField);
    _editorHandler.showEditor();
  }
  
  /**
   * Closes the currently open editor.
   *
   * @param doCommitIfValid Denotes if a possibly valid value should be committed before closing.
   */
  public void requestCloseEditor(boolean doCommitIfValid)
  {
    if ( _editorHandler != null && _editorHandler.isCurrentlyOpen() )
      return;
    _editorHandler.closeEditor(doCommitIfValid);
  }
  
  /********************************** INTERNAL LOGIC *****************************************/
  
  private Node createCellComponents()
  {
    GridPane grid = new GridPane();
    grid.setPadding(new Insets(0, 10, 0, 2));
    grid.setAlignment(Pos.CENTER);
    //    grid.addEventFilter(MouseEvent.MOUSE_PRESSED, event -> {
    //      if ( _editorHandler != null && _editorHandler.isCurrentlyOpen() )
    //        event.consume();
    //    });
    
    // Set up the columns' growing/shrinking behaviour
    ColumnConstraints col1 = new ColumnConstraints();
    ColumnConstraints col2 = new ColumnConstraints();
    col1.setPercentWidth(2);
    col2.setPercentWidth(46);
    col2.setMaxWidth(120);
    col2.setMinWidth(40);
    grid.getColumnConstraints().addAll(col1, col2, col1, col1, col2, col1);
    
    // Build the two panels holding the values
    _minValuePane = new StackPane();
    _maxValuePane = new StackPane();
    grid.add(_minValuePane, 1, 0);
    grid.add(_maxValuePane, 4, 0);
    _minValuePane.setAlignment(Pos.CENTER);
    _maxValuePane.setAlignment(Pos.CENTER);
    //    minValuePane.setStyle("-fx-background-color: red;");
    //    maxValuePane.setStyle("-fx-background-color: red;");
    
    // Now, add controls to the panels
    _valueLabel1 = new Label();
    _valueLabel2 = new Label();
    _minValuePane.getChildren().add(_valueLabel1);
    _maxValuePane.getChildren().add(_valueLabel2);
    
    // Fill the labels with according text. A call to "updateLabels()" is being circumvented in this case due to circular call exception
    _valueLabel1.setText(_valueConverter.toString(WhichField.MIN, _value1, DEFAULT_STRING_FORMAT));
    _valueLabel2.setText(_valueConverter.toString(WhichField.MAX, _value2, DEFAULT_STRING_FORMAT));
    
    // Apply double-click listeners
    _valueLabel1.addEventFilter(MouseEvent.MOUSE_PRESSED, (mouseEvent) -> {
      if ( _isEditable && mouseEvent.getButton().equals(MouseButton.PRIMARY) && mouseEvent.getClickCount() == 2 )
      {
        requestOpenEditor(WhichField.MIN);
        mouseEvent.consume();
      }
    });
    _valueLabel2.addEventFilter(MouseEvent.MOUSE_PRESSED, (mouseEvent) -> {
      if ( _isEditable && mouseEvent.getButton().equals(MouseButton.PRIMARY) && mouseEvent.getClickCount() == 2 )
      {
        requestOpenEditor(WhichField.MAX);
        mouseEvent.consume();
      }
    });
    
    //    grid.setOnKeyPressed(event -> {          ----> Is not being called
    //      if ( event.getCode() == KeyCode.F2 )
    //      {
    //        requestOpenEditor(WhichField.MIN);
    //        event.consume();
    //      }
    //    });
    return grid;
  }
  
  /*******************************************************************************************/
  
  public static final ValueConverter DEFAULT_CONVERTER = new DefaultDoubleConverter();
  
  private static class DefaultDoubleConverter implements ValueConverter
  {
    
    /** {@inheritDoc} */
    @Override
    public String toString(WhichField whichField, double value, String format)
    {
      try
      {
        if ( Double.isNaN(value) )
          return "";
        else
          return new DecimalFormat(format, DecimalFormatSymbols.getInstance(Locale.ENGLISH)).format(value);
      }
      catch (Exception ex)
      {
        return "";
      }
    }
    
    /** {@inheritDoc} */
    @Override
    public Double convertValue(WhichField whichField, String valueString)
    {
      if ( valueString == null || valueString.isEmpty() )
        return null;
      else if ( "NaN".equals(valueString) )//$NON-NLS-1$
        return Double.NaN;
      else
      {
        try
        {
          return Double.parseDouble(valueString);
        }
        catch (NumberFormatException e)
        {
          return null;
        }
      }
    }
    
  } /* class DoubleConverter */
  
  
  
  /*******************************************************************************************/
  
  private class EditorHandler
  {
    private StackPane  _container;
    private double     _initialValue;
    private WhichField _whichField;
    
    private TextField               _textField;
    private final ValidationSupport _validationSupport = new ValidationSupport();
    
    private boolean _isCurrentlyOpen;
    private Node    _previousFocusOwner = null;
    
    public EditorHandler(StackPane container, double initialValue, WhichField whichField)
    {
      _container = container;
      _initialValue = initialValue;
      _whichField = whichField;
      _isCurrentlyOpen = false;
      
      _textField = new TextField() {
        @Override
        public void insertText(int index, String text)
        {
          String fixedText = fixText(text);
          super.insertText(index, fixedText);
        }
        
        @Override
        public void replaceText(int start, int end, String text)
        {
          String fixedText = fixText(text);
          if ( StringUtil.isEmpty(text) || !StringUtil.isEmpty(fixedText) ) // The 'StringUtil.isEmpty(text)' condition is necessary to make text erasable by backspace
            super.replaceText(start, end, fixedText);
        }
        
        @Override
        public void replaceText(IndexRange range, String text)
        {
          String fixedText = fixText(text);
          if ( StringUtil.isEmpty(text) || !StringUtil.isEmpty(fixedText) ) // The 'StringUtil.isEmpty(text)' condition is necessary to make text erasable by backspace
            replaceText(range.getStart(), range.getEnd(), fixedText);
        }
        
        private String fixText(String text)
        {
          final char allowedDecimalSeparator = '.';
          text = text.replace(' ', '\u00a0');//$NON-NLS-1$
          text = text.replace(',', allowedDecimalSeparator);//$NON-NLS-1$
          
          String retVal = "";
          final Set<Character> allowedSymbols = (_allowedExtraSymbols != null) ? _allowedExtraSymbols : new HashSet<Character>();
          for ( int strPos = 0; strPos < text.length(); strPos++ )
          {
            char c = text.charAt(strPos);
            if ( c == allowedDecimalSeparator || allowedSymbols.contains(new Character(c)) || (c >= '0' && c <= '9') )
              retVal += c;
          }
          return retVal;
        }
      };
      if ( _isHighlightedIfInvalid )
      {
        _validationSupport.setValidationDecorator(new StyleClassValidationDecoration());
        _validationSupport.registerValidator(_textField, false, Validator.createPredicateValidator((String a) -> tryParsing(), "", Severity.ERROR));
      }
    }
    
    public boolean isCurrentlyOpen()
    {
      return _isCurrentlyOpen;
    }
    
    public void showEditor()
    {
      if ( _isCurrentlyOpen )
        return;
      
      _previousFocusOwner = _container.getScene().getFocusOwner();
      
      _isCurrentlyOpen = true;
      _textField.setText(_valueConverter.toString(_whichField, _initialValue, getCell().getFormat()));
      _container.getChildren().add(_textField);
      _textField.setMaxHeight(8);
      _textField.setPadding(new Insets(1));
      _textField.setAlignment(Pos.CENTER);
      _textField.requestFocus();
      _textField.requestFocus(); // --> This "hack" keeps the TextField from losing the first key press under certain circumstances.
      _textField.selectAll();
      
      attachEnterEscapeEventHandler();
      _textField.focusedProperty().addListener(FocusLostListener);
    }
    
    public void closeEditor(boolean doCommitIfValid)
    {
      if ( !_isCurrentlyOpen )
        return;
      
      _isCurrentlyOpen = false;
      _textField.setOnKeyPressed(null);
      _textField.focusedProperty().removeListener(FocusLostListener);
      _container.getChildren().remove(_textField);
      
      if ( _previousFocusOwner != null )
        _previousFocusOwner.requestFocus();
      
      if ( doCommitIfValid && tryParsing() )
      {
        
        Double value = _valueConverter.convertValue(_whichField, _textField.getText());
        if ( _commitValueCallback != null )
          _commitValueCallback.commitValue(getCell(), _whichField, value, _textField.getText());
        
        if ( _whichField == WhichField.MIN )
          _value1 = value.doubleValue();
        else if ( _whichField == WhichField.MAX )
          _value2 = value.doubleValue();
        
        updateLabels();
      }
    }
    
    private void attachEnterEscapeEventHandler()
    {
      _textField.setOnKeyPressed(new EventHandler<KeyEvent>() {
        @Override
        public void handle(KeyEvent t)
        {
          if ( t.getCode() == KeyCode.ENTER )
          {
            if ( "".equals(_textField.getText()) ) //$NON-NLS-1$
              closeEditor(false);
            else if ( tryParsing() == true )
              closeEditor(true);
            if ( _whichField == WhichField.MIN )
              requestOpenEditor(WhichField.MAX);
            t.consume();
          }
          else if ( t.getCode() == KeyCode.ESCAPE )
          {
            closeEditor(false);
            t.consume();
          }
          else if ( t.getCode() == KeyCode.TAB )
          {
            closeEditor(true);
            t.consume();
            if ( _whichField == WhichField.MIN /*&& t.isShiftDown() == false*/ )
              requestOpenEditor(WhichField.MAX);
            if ( _whichField == WhichField.MAX /*&& t.isShiftDown() == true*/ )
              requestOpenEditor(WhichField.MIN);
          }
        }
      });
      
    }
    
    private Boolean tryParsing()
    {
      if ( _valueConverter.convertValue(_whichField, _textField.getText()) == null )
        return false;
      if ( _checkValueCallback != null )
        if ( !_checkValueCallback.apply(_whichField, _textField.getText()) )
          return false;
        
      return true;
    }
    
    
    private final ChangeListener<Boolean> FocusLostListener = new ChangeListener<Boolean>() {
      @Override
      public void changed(ObservableValue<? extends Boolean> observable, Boolean oldValue, Boolean newValue)
      {
        if ( !newValue )
          closeEditor(true);
        else
          closeEditor(false);
      }
    };
    
    
    
  }
  
  
  /*******************************************************************************************/
  
  
  /*******************************************************************************************/
  
  /**
   * Just copied from {@link SpreadsheetCellType} and modified a bit.
   */
  public class MinMaxDoubleCellType extends SpreadsheetCellType<Object>
  {
    public MinMaxDoubleCellType()
    {
      this(new StringConverterWithFormat<Object>() {
        @Override
        public Object fromString(String arg0)
        {
          return arg0;
        }
        
        @Override
        public String toString(Object arg0)
        {
          return arg0 == null ? "" : arg0.toString(); //$NON-NLS-1$
        }
      });
    }
    
    public MinMaxDoubleCellType(StringConverterWithFormat<Object> converter)
    {
      super(converter);
    }
    
    @Override
    public String toString()
    {
      return "object"; //$NON-NLS-1$
    }
    
    @Override
    public boolean match(Object value)
    {
      return true;
    }
    
    /**
     * Creates a cell that hold an Object at the specified position, with the specified row/column span.
     * 
     * @param row row number
     * @param column column number
     * @param rowSpan rowSpan (1 is normal)
     * @param columnSpan ColumnSpan (1 is normal)
     * @param value the value to display
     * @return a {@link SpreadsheetCell}
     */
    public SpreadsheetCell createCell(final int row, final int column, final int rowSpan, final int columnSpan, final Object value)
    {
      SpreadsheetCell cell = new SpreadsheetCellBase(row, column, rowSpan, columnSpan, this);
      cell.setItem(value);
      return cell;
    }
    
    @Override
    public SpreadsheetCellEditor createEditor(SpreadsheetView view)
    {
      requestOpenEditor(WhichField.MIN);
      return null;
    }
    
    /**
     * TODO {@inheritDoc}
     *
     * @see org.controlsfx.control.spreadsheet.SpreadsheetCellType#toString(java.lang.Object)
     */
    @Override
    public String toString(Object object)
    {
      return "";
    }
    
    /**
     * TODO {@inheritDoc}
     *
     * @see org.controlsfx.control.spreadsheet.SpreadsheetCellType#convertValue(java.lang.Object)
     */
    @Override
    public Object convertValue(Object value)
    {
      return "";
    }
    
    
  };
  
}
