package zorg.gridviewdemo;

import javafx.application.Application;
import javafx.application.Platform;
import javafx.beans.property.IntegerProperty;
import javafx.beans.property.SimpleIntegerProperty;
import javafx.scene.Scene;
import javafx.scene.control.Label;
import javafx.scene.control.ListCell;
import javafx.scene.control.ListView;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.VBox;
import javafx.stage.Stage;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.function.Function;

import static java.util.Objects.requireNonNull;

public class ListViewDemo extends Application {

    public static void main(String[] args) {
        launch(args);
    }

    private final ExecutorService executor = Executors.newSingleThreadExecutor();
    private final Map<Path,CompletableFuture<Image>> cache = new HashMap<>();

    private final IntegerProperty updateItemCount = new SimpleIntegerProperty(0);
    private final Label updateItemCountLabel = new Label();

    @Override
    public void start(Stage primaryStage) throws Exception {

        // TODO edit to point to suitable folder containing jpgs
        Path jpgFolder = Paths.get("C:\\Users\\Public\\Pictures\\Sample Pictures");

        updateItemCountLabel.setMaxWidth(Double.POSITIVE_INFINITY);
        updateItemCountLabel.textProperty().bind(updateItemCount.asString("Update Item Counter: %d"));

        BorderPane root = new BorderPane();
        ListView<FileInfo> listView = new ListView<>();
        listView.setFixedCellSize(180);

        listView.setCellFactory(gv -> new ImageCell());
        root.setCenter(listView);
        root.setBottom(updateItemCountLabel);
        root.setPrefSize(800, 600);
        Scene scene = new Scene(root);
        primaryStage.setScene(scene);
        primaryStage.show();

        CompletableFuture.supplyAsync(() -> {
            List<FileInfo> fileInfos = new ArrayList<>();
            try {
                for (Path path : Files.newDirectoryStream(jpgFolder, "*.jpg")) {
                    fileInfos.add(new FileInfo(path));
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
            return fileInfos;
        }).thenAcceptAsync(fileInfos -> listView.getItems().setAll(fileInfos), Platform::runLater);
    }

    @Override
    public void stop() throws Exception {
        executor.shutdown();
    }

    private CompletableFuture<Image> requestImage(Path path) {
        return cache.computeIfAbsent(path,
                p -> CompletableFuture.supplyAsync(() -> {
                    // artificial delay for demo purposes
                    try {Thread.sleep(500);}catch(Exception e) {}
                    return new Image(p.toUri().toASCIIString(), 200, 200, true, true, false);
                }, executor)
                // the cached future must be on the Platform thread so we can use the non-async thenAccept method
                // which will execute immediately once the future is complete, reducing the number of Platform::runLater
                // tasks that would otherwise be created by the ImageCell.
                .thenApplyAsync(Function.identity(), Platform::runLater)
        );
    }

    public class ImageCell extends ListCell<FileInfo> {
        private final ImageView imageView = new ImageView();
        private final Label label = new Label();

        public ImageCell() {
            VBox vbox = new VBox();
            vbox.getChildren().addAll(imageView, label);
            setGraphic(vbox);
        }

        @Override
        protected void updateItem(FileInfo item, boolean empty) {
//            System.out.println("updateItem(" + item + ", " + empty + ")");
            updateItemCount.set(updateItemCount.get() + 1);
            super.updateItem(item, empty);
            if (empty) {
                imageView.setImage(null);
                label.setText("");
            } else {
                label.setText(item.path.getFileName().toString());
                imageView.setImage(null);
                requestImage(item.path).thenAccept((value) -> {
                    if (item.equals(getItem())) {
                        imageView.setImage(value);
                    }
                });
            }
        }
    }

    public static class FileInfo {
        public final Path path;

        public FileInfo(Path path) {
            this.path = requireNonNull(path);
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            FileInfo fileInfo = (FileInfo) o;

            return path.equals(fileInfo.path);
        }

        @Override
        public int hashCode() {
            return path.hashCode();
        }

        @Override
        public String toString() {
            return "FileInfo[" + path + "]";
        }
    }
}
