package org.clemens.expandable;

import javafx.animation.KeyFrame;
import javafx.animation.KeyValue;
import javafx.animation.Timeline;
import javafx.beans.property.DoubleProperty;
import javafx.beans.property.SimpleDoubleProperty;
import javafx.geometry.Pos;
import javafx.scene.control.SkinBase;
import javafx.scene.layout.StackPane;
import javafx.scene.shape.Rectangle;
import javafx.util.Duration;
import org.clemens.expandable.ExpandablePane;

import javax.annotation.Nonnull;

import static java.util.Objects.requireNonNull;

public class ExpandablePaneSkin extends SkinBase<ExpandablePane>
{
	private static final double SCALE_EXPANDED = 1d;
	private static final double SCALE_COLLAPSED = 0d;

	@Nonnull private final TitleBar title = new TitleBar();

	@Nonnull private final ExpandablePane control;

	@Nonnull private final StackPane content = new StackPane();
	@Nonnull private final Rectangle clip = new Rectangle();

	@Nonnull private final DoubleProperty scale = new SimpleDoubleProperty(SCALE_EXPANDED);
	@Nonnull private final Timeline animation = new Timeline();
	@Nonnull private double scaleTarget = SCALE_EXPANDED;

	public ExpandablePaneSkin(@Nonnull final ExpandablePane control)
	{
		super(control);
		this.control = requireNonNull(control);

		getChildren().addAll(title, content);
		handlePositionChange(control.getPosition());

		content.setClip(clip);

		if (control.getContent() != null)
			content.getChildren().add(control.getContent());

		control.contentProperty().addListener((ov, o, n) -> {
			if (n == null)
				content.getChildren().clear();
			else
				content.getChildren().setAll(n);
		});

		title.titleProperty().bind(control.titleProperty());
		title.graphicProperty().bind(control.graphicProperty());
		title.positionProperty().bind(control.positionProperty());
		title.addEventHandler(ExpandEvent.INVERT, event -> collapseAndExpand());

		scale.addListener(observable -> control.requestLayout());
		control.positionProperty().addListener((ov, o, n) -> handlePositionChange(n));
		control.expandedProperty().addListener((ov, o, expanded) -> collapseAndExpandNoAnimation(expanded));

		collapseAndExpandNoAnimation(control.isExpanded());
	}

	private void handlePositionChange(@Nonnull final Position pos)
	{
		if (pos == Position.TOP || pos == Position.BOTTOM)
		{
			clip.scaleXProperty().unbind();
			clip.scaleYProperty().bind(scale);
			clip.setScaleX(SCALE_EXPANDED);

			content.setAlignment(pos == Position.BOTTOM ? Pos.BOTTOM_LEFT : Pos.TOP_LEFT);
		}
		else
		{
			clip.scaleYProperty().unbind();
			clip.scaleXProperty().bind(scale);
			clip.setScaleY(SCALE_EXPANDED);

			content.setAlignment(pos == Position.LEFT ? Pos.TOP_LEFT : Pos.TOP_RIGHT);
		}
	}

	private void collapseAndExpand()
	{
		animation.stop();

		scaleTarget = scaleTarget < SCALE_EXPANDED ? SCALE_EXPANDED : SCALE_COLLAPSED;

		final KeyValue kv = new KeyValue(scale, scaleTarget);
		final KeyFrame kf = new KeyFrame(Duration.millis(control.getAnimationDuration()), kv);

		animation.getKeyFrames().setAll(kf);
		animation.play();

		title.setExpanded(scaleTarget > SCALE_COLLAPSED);
	}

	private void collapseAndExpandNoAnimation(final boolean expanded)
	{
		animation.stop();
		scaleTarget = expanded ? SCALE_EXPANDED : SCALE_COLLAPSED;
		scale.setValue(scaleTarget);
		title.setExpanded(expanded);
	}

	@Override
	protected void layoutChildren(
			final double contentX,
			final double contentY,
			final double contentWidth,
			final double contentHeight)
	{
		final Position pos = control.getPosition();

		if (pos == Position.BOTTOM || pos == Position.TOP)
		{
			final double titlePaneHeight = title.minHeight(-1);

			if (pos == Position.TOP)
			{
				title.relocate(0, 0);
				content.relocate(0, titlePaneHeight);
			}
			else
			{
				title.relocate(0, contentHeight - titlePaneHeight);
				content.relocate(0, 0);
			}

			title.resize(contentWidth, titlePaneHeight);
			content.resize(contentWidth, contentHeight - titlePaneHeight);

			clip.setWidth(contentWidth);
			clip.setHeight(content.prefHeight(-1));
		}
		else
		{
			final double titlePaneWidth = title.minWidth(-1);

			if (pos == Position.LEFT)
			{
				title.relocate(0, 0);
				content.relocate(titlePaneWidth, 0);
			}
			else
			{
				title.relocate(contentWidth - titlePaneWidth, 0);
				content.relocate(0, 0);
			}

			final double availableWidth = contentWidth - titlePaneWidth;
			title.resize(titlePaneWidth, contentHeight);
			content.resize(availableWidth, contentHeight);

			clip.setWidth(Math.max(content.prefWidth(-1), availableWidth));
			clip.setHeight(contentHeight);
		}
	}

	@Override
	protected double computePrefHeight(
			final double width,
			final double topInset,
			final double rightInset,
			final double bottomInset,
			final double leftInset)
	{
		if (control.getPosition() == Position.BOTTOM || control.getPosition() == Position.TOP)
			return title.prefHeight(width) + topInset + bottomInset + snapSize(content.prefHeight(width) * scale.getValue());
		else
			return Math.max(title.prefHeight(width), content.prefHeight(width)) + topInset + bottomInset;
	}

	@Override
	protected double computeMinHeight(
			final double width,
			final double topInset,
			final double rightInset,
			final double bottomInset,
			final double leftInset)
	{
		if (control.getPosition() == Position.BOTTOM || control.getPosition() == Position.TOP)
			return title.minHeight(width) + topInset + bottomInset + snapSize(content.minHeight(width) * scale.getValue());
		else
			return Math.max(title.minHeight(width), content.minHeight(width)) + topInset + bottomInset;
	}

	@Override
	protected double computePrefWidth(
			final double height,
			final double topInset,
			final double rightInset,
			final double bottomInset,
			final double leftInset)
	{
		if (control.getPosition() == Position.LEFT || control.getPosition() == Position.RIGHT)
			return title.prefWidth(height) + leftInset + rightInset + snapSize(content.prefWidth(height) * scale.getValue());
		else
			return Math.max(title.prefWidth(height), content.prefWidth(height)) + leftInset + rightInset;
	}

	@Override
	protected double computeMinWidth(
			final double height,
			final double topInset,
			final double rightInset,
			final double bottomInset,
			final double leftInset)
	{
		if (control.getPosition() == Position.LEFT || control.getPosition() == Position.RIGHT)
			return title.minWidth(height) + leftInset + rightInset + snapSize(content.minWidth(height) * scale.getValue());
		else
			return Math.max(title.minWidth(height), content.minWidth(height)) + leftInset + rightInset;
	}

}
