package org.clemens.expandable;

import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.scene.Group;
import javafx.scene.Node;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

public class RotatedNode extends Group
{
	public enum Amount
	{
		NONE(0),
		ONE_QUARTER(90),
		TWO_QUARTER(180),
		THREE_QUARTER(270);

		private final int value;

		Amount(final int value)
		{
			this.value = value;
		}
	}

	private static final double NOT_SET = Double.NEGATIVE_INFINITY;

	private double width = NOT_SET;
	private double height = NOT_SET;

	@Nonnull private final ObjectProperty<Node> content = new SimpleObjectProperty<>(null);
	@Nonnull private final ObjectProperty<Amount> amount = new NonNullObjectProperty<>(Amount.NONE);

	public RotatedNode(@Nullable final Node content)
	{
		this(content, Amount.NONE);
	}

	public RotatedNode(@Nullable final Node content, @Nonnull final Amount amount)
	{
		this.amount.addListener(observable -> updateContent());
		this.content.addListener((ov, o, n) -> updateContent());

		this.content.set(content);
		this.amount.set(amount);
	}

	private void updateContent()
	{
		width = NOT_SET;
		height = NOT_SET;

		if (content.get() != null)
		{
			content.get().setRotate(amount.get().value);
			super.getChildren().setAll(content.get());
		}
		else
			super.getChildren().clear();

		requestLayout();
	}

	@Override
	public ObservableList<Node> getChildren()
	{
		return FXCollections.unmodifiableObservableList(super.getChildren());
	}

	@Override
	protected void layoutChildren()
	{
		if (content.get() != null)
		{
			if (width == NOT_SET || height == NOT_SET)
			{
				content.get().autosize();
			}
			else
			{
				final double w;
				final double h;

				if (swapHeightAndWidth())
				{
					w = height;
					h = width;
				}
				else
				{
					w = width;
					h = height;
				}

				content.get().resize(w, h);
			}
		}
	}

	@Override
	public boolean isResizable()
	{
		return true;
	}

	@Override
	public void resize(final double width, final double height)
	{
		this.width = width;
		this.height = height;

		layoutChildren();
	}

	@Override
	public double prefWidth(final double height)
	{
		if (width == NOT_SET)
			return super.prefWidth(-1);
		else
		{
			final Node node = content.get();
			if (node == null)
				return 0d;
			else
				return swapHeightAndWidth() ? node.prefHeight(-1) : node.prefWidth(-1);
		}
	}

	@Override
	public double prefHeight(final double width)
	{
		if (height == NOT_SET)
			return super.prefHeight(-1);
		else
		{
			final Node node = content.get();
			if (node == null)
				return 0d;
			else
				return swapHeightAndWidth() ? node.prefWidth(-1) : node.prefHeight(-1);
		}
	}

	@Override
	public double minHeight(final double width)
	{
		final Node node = content.get();
		return node == null ? 0d : swapHeightAndWidth() ? node.minWidth(-1) : node.minHeight(-1);
	}

	@Override
	public double minWidth(final double height)
	{
		final Node node = content.get();
		return node == null ? 0d : swapHeightAndWidth() ? node.minHeight(-1) : node.minWidth(-1);
	}

	private boolean swapHeightAndWidth()
	{
		return amount.get() == Amount.ONE_QUARTER || amount.get() == Amount.THREE_QUARTER;
	}

	@Nonnull
	public Node getContent()
	{
		return content.get();
	}

	@Nonnull
	public ObjectProperty<Node> contentProperty()
	{
		return content;
	}

	public void setContent(@Nullable final Node content)
	{
		this.content.set(content);
	}

	@Nonnull
	public Amount getAmount()
	{
		return amount.get();
	}

	@Nonnull
	public ObjectProperty<Amount> amountProperty()
	{
		return amount;
	}

	public void setAmount(@Nonnull final Amount amount)
	{
		this.amount.set(amount);
	}
}
