package org.clemens.expandable;

import javafx.event.ActionEvent;
import javafx.scene.Node;
import javafx.scene.control.Label;
import javafx.scene.control.SkinBase;
import javafx.scene.input.MouseEvent;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.Objects;
import java.util.Optional;
import java.util.function.BiFunction;
import java.util.function.Predicate;


public class TitleBarSkin extends SkinBase<TitleBar>
{
	private static final String STYLE_CLASS = "title-bar";
	private static final String LABEL_STYLE_CLASS = "title-label";

	@Nonnull private final ExpanderButton expander = new ExpanderButton();
	@Nonnull private final Label title = new Label();
	@Nonnull private final RotatedNode titleNode = new RotatedNode(title, RotatedNode.Amount.NONE);
	@Nonnull private Optional<RotatedNode> graphic = Optional.empty();

	@Nonnull private final TitleBar control;

	public TitleBarSkin(@Nonnull final TitleBar control)
	{
		super(control);
		this.control = Objects.requireNonNull(control);

		control.getStyleClass().add(STYLE_CLASS);
		title.getStyleClass().add(LABEL_STYLE_CLASS);

		getChildren().addAll(expander, titleNode);

		title.textProperty().bind(control.titleProperty());

		graphic = updateGraphic(control.getGraphic());
		control.graphicProperty().addListener((ov, o, n) -> {
			graphic = updateGraphic(n);
		});

		titleNode.setAmount(control.getPosition().getRotationAmount());
		control.getPosition().configureExpanderButton(expander);

		control.positionProperty().addListener((ov, o, n) -> {
			titleNode.setAmount(n.getRotationAmount());
			graphic.ifPresent(g -> g.setAmount(n.getRotationAmount()));
			n.configureExpanderButton(expander);
		});

		expander.setExpanded(control.getExpanded());
		control.expandedProperty().addListener((ov, o, expanded) -> expander.setExpanded(expanded));

		expander.addEventFilter(ActionEvent.ANY, event -> {
			event.consume();
			control.fireEvent(new ExpandEvent());
		});
		control.addEventHandler(MouseEvent.MOUSE_PRESSED, event -> control.fireEvent(new ExpandEvent()));
	}

	@Nonnull
	private Optional<RotatedNode> updateGraphic(@Nullable final Node graphic)
	{
		this.graphic.ifPresent(g -> getChildren().remove(g));

		if (graphic == null)
			return Optional.empty();
		else
		{
			final RotatedNode graphicNode = new RotatedNode(graphic, control.getPosition().getRotationAmount());
			getChildren().add(graphicNode);
			return Optional.of(graphicNode);
		}
	}

	@Override
	protected void layoutChildren(
			final double contentX,
			final double contentY,
			final double contentWidth,
			final double contentHeight)
	{
		if (TitleBar.isHorizontalPosition(control.getPosition()))
			positionChildren(
					contentX,
					contentY,
					contentWidth,
					contentHeight,
					Node::minWidth,
					Node::prefWidth,
					Node::resize,
					Node::relocate
			);
		else
			positionChildren(
					contentY,
					contentX,
					contentHeight,
					contentWidth,
					Node::minHeight,
					Node::prefHeight,
					(n, h, w) -> n.resize(w, h),
					(n, y, x) -> n.relocate(x, y)
			);
	}

	@SuppressWarnings("MethodWithTooManyParameters")
	private void positionChildren(
			final double contentMainPos,
			final double contentMinorPos,
			final double contentMainSize,
			final double contentMinorSize,
			@Nonnull final BiFunction<Node, Double, Double> minSize,
			@Nonnull final BiFunction<Node, Double, Double> prefSize,
			@Nonnull final TriConsumer<Node, Double, Double> resize,
			@Nonnull final TriConsumer<Node, Double, Double> relocate)
	{
		final double spacing = control.getSpacing();
		final double neededForSpacing = graphic.map(g -> 2 * spacing).orElse(spacing);

		final double availableSize = contentMainSize - neededForSpacing;
		final double expanderSize = minSize.apply(expander, -1d);

		double delta = contentMainPos;

		resize.accept(expander, expanderSize, contentMinorSize);
		relocate.accept(expander, delta, contentMinorPos);
		delta += expanderSize + spacing;

		final double prefGraphicSize = graphic.map(g -> prefSize.apply(g, -1d)).orElse(0d);
		final double availableTitleSize = availableSize - expanderSize - prefGraphicSize;
		final double titleSize = Math.min(Math.max(availableTitleSize, minSize.apply(titleNode, -1d)),
				prefSize.apply(titleNode, -1d));

		resize.accept(titleNode, titleSize, contentMinorSize);
		relocate.accept(titleNode, delta, contentMinorPos);
		delta += titleSize + spacing;
		final double effectivleyFinalDeltaX = delta;

		graphic.ifPresent(g -> {
			final double availabeGraphicSize = availableSize - expanderSize - titleSize;
			resize.accept(g, availabeGraphicSize, contentMinorSize);
			relocate.accept(g, effectivleyFinalDeltaX, contentMinorPos);
		});
	}

	@Override
	protected double computeMinWidth(
			final double height,
			final double topInset,
			final double rightInset,
			final double bottomInset,
			final double leftInset)
	{
		return compSize(
				Node::minWidth,
				TitleBar::isHorizontalPosition,
				height,
				leftInset,
				rightInset
		);
	}

	@Override
	protected double computeMinHeight(
			final double width,
			final double topInset,
			final double rightInset,
			final double bottomInset,
			final double leftInset)
	{
		return compSize(
				Node::minHeight,
				TitleBar::isVerticalPosition,
				width,
				topInset,
				bottomInset
		);
	}

	@Override
	protected double computePrefWidth(
			final double height,
			final double topInset,
			final double rightInset,
			final double bottomInset,
			final double leftInset)
	{
		return compSize(
				Node::prefWidth,
				TitleBar::isHorizontalPosition,
				height,
				leftInset,
				rightInset
		);
	}

	@Override
	protected double computePrefHeight(
			final double width,
			final double topInset,
			final double rightInset,
			final double bottomInset,
			final double leftInset)
	{
		return compSize(
				Node::prefHeight,
				TitleBar::isVerticalPosition,
				width,
				topInset,
				bottomInset
		);
	}

	@SuppressWarnings("MethodWithTooManyParameters")
	private double compSize(
			@Nonnull final BiFunction<Node, Double, Double> sizeFunction,
			@Nonnull final Predicate<Position> isMainOrientation,
			final double restricedSize,
			final double insetsA,
			final double insetsB)
	{
		final double expanderSize = sizeFunction.apply(expander, restricedSize);
		final double titleSize = sizeFunction.apply(titleNode, restricedSize);
		final double graphicSize = graphic.map(g -> sizeFunction.apply(g, restricedSize)).orElse(0d);

		final double insets = insetsA + insetsB;

		if (isMainOrientation.test(control.getPosition()))
		{
			final double spacing = graphic.map(g -> 2 * control.getSpacing()).orElse(control.getSpacing());

			return expanderSize + titleSize + graphicSize
					+ insets + spacing;
		}
		else
		{
			final double controlsWidth = Math.max(Math.max(expanderSize, titleSize), graphicSize);
			return controlsWidth + insets;
		}
	}

}
